#!/usr/bin/env python3

# SPDX-License-Identifier: LGPL-2.1-or-later

# Copyright (C) 2021 Pengutronix, Jonas Martin <entwicklung@pengutronix.de>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

import subprocess
from time import sleep
from .usbmuxctl import Mux
from .usbmuxctl import path_from_usb_dev
import usb.core
from enum import Enum
import struct
import logging
import argparse

try:
    import coloredlogs
    coloredlogs.install()
except:
    pass

# Vendor/product ID for STM32 DFU bootloader
DFU_VENDOR_ID=0x0483
DFU_PRODUCT_ID=0xdf11

DFU_UTIL_CMD = "dfu-util"

class _DFUComand(Enum):
    DFU_DETACH = 0
    DFU_DNLOAD = 1
    DFU_UPLOAD = 2
    DFU_GETSTATUS = 3
    DFU_CLRSTATUS = 4
    DFU_GETSTATE = 5
    DFU_ABORT = 6

class _bStatus(Enum):
    OK = 0x00
    errTARGET = 0x01
    errFILE = 0x02
    errWRITE = 0x03
    errERASE = 0x04
    errCHECK_ERASED = 0x05
    errPROG = 0x06
    errVERIFY = 0x07
    errADDRESS = 0x08
    errNOTDONE = 0x09
    errFIRMWARE = 0x0a
    errVENDOR = 0x0b
    errUSBR = 0x0c
    errPOR = 0x0d
    errUNKNOWN = 0xe
    errSTALLEDPKT = 0x0f

class _bState(Enum):
    appIDLE = 0
    appDETACH = 1
    dfuIDLE = 2
    dfuDNLOAD_SYNC = 3
    dfuDNBUSY = 4
    dfuDNLOAD_IDLE = 5
    dfuMANIFEST_SYNC = 6
    dfuMANIFEST = 7
    dfuMANIFEST_WAIT_RESET = 8
    dfuUPLOAD_IDLE = 9
    dfuERROR = 10


class NoDFUDeviceFound(IOError):
    pass
class ToManyDFUDeviceFound(IOError):
    def __init__(self, paths):
        self.paths = paths


class DFUException(Exception):
    def __init__(self, status):
        self.status = status

class DFU:
    """STM32 DFU implementation.
    Only to get out of DFU mode"""

    # See RM0091 32.4.1 MCU device ID code
    STM32_DEVICES = {
            0x444: "STM32F03x",
            0x445: "STM32F04x",
            0x440: "STM32F05x",
            0x448: "STM32F07x",
            0x442: "STM32F09x",
            }

    STM32_VERSIONS = {
            0x1000: "1.0",
            0x1001: "1.1",
            0x2000: "2.0",
            0x2001: "2.1",
            }


    def __init__(self, path):
        """Create an DFU instance.

        Can be used to read memory and exit the DFU bootloader.

        Arguments:
          path -- USB path to the USB device (example: 1-2.2.1)

        Exceptions:
          NoDFUDeviceFound -- Could not find a device in DFU mode at the path.
          ToManyDFUDeviceFound -- Found more then one device. Unclear which to choose.
        """
        def find_filter(dev):
            if not path is None:
                dev_path = path_from_usb_dev(dev)
                return dev_path == path
            return True
        dev = usb.core.find(idVendor=DFU_VENDOR_ID, idProduct=DFU_PRODUCT_ID, custom_match=find_filter, find_all=True)

        # Make sure we got one device
        if dev is None:
            raise NoDFUDeviceFound()

        dev = list(dev)
        if len(dev) == 0:
            raise NoDFUDeviceFound()

        if len(dev) > 1:
            paths = []
            for d in dev:
                path = path_from_usb_dev(d)
                paths.append(path)
            raise ToManyDFUDeviceFound(paths)

        self._dev = dev[0]

        self.log = logging.getLogger("DFU")

        self.interface = 0

        self._dev.set_configuration()

    @staticmethod
    def parse_dfu_status(status):
        """Generate a human readable dict from DFU_STATUS response"""
        out = {}
        out["bStatus"] = _bStatus(status[0])
        out["bState"] = _bState(status[4])
        bwPollTimeout = status[1] | (status[2]<<8) | (status[3]<<16)
        out["bwPollTimeout"] = bwPollTimeout
        return out



    def get_path(self):
        return path_from_usb_dev(self._dev)

    def _cmd_out(self, cmd, wValue, data=None, bmRequestType=0x21):
        """Host to device"""
        ret = self._dev.ctrl_transfer(
                bmRequestType = bmRequestType,
                bRequest = cmd.value,
                wValue = wValue,
                wIndex = self.interface,
                data_or_wLength = data)
        return ret

    def _cmd_in(self, cmd, wValue, data=None, bmRequestType=0xa1):
        """Device to Host"""
        ret = self._dev.ctrl_transfer(
                bmRequestType = bmRequestType,
                bRequest = cmd.value,
                wValue = wValue,
                wIndex = self.interface,
                data_or_wLength = data)
        return ret

    def _get_status(self):
        """DFU internal command
        Get the device status"""
        ret = self._cmd_in(_DFUComand.DFU_GETSTATUS, 0, 6)
        ret = self.parse_dfu_status(ret)
        self.log.debug("get_status: %s", ret)
        return ret

    def _check_status(self):
        """Wrapper for _get_status.
        Returns the device status and thows DFUException if bStatus is not OK"""

        status = self._get_status()
        if status["bStatus"] != _bStatus.OK:
            raise DFUException(status)
        return status

    def _clear_status(self):
        """DFU internal command
        Clears status and possible error states"""

        self.log.debug("Clear status")
        ret = self._cmd_out(_DFUComand.DFU_CLRSTATUS, 0)
        self._check_status()

    def _abort(self):
        """DFU internal command
        Abort currently running command"""
        self.log.debug("Abort")
        ret = self._cmd_out(_DFUComand.DFU_ABORT, 0)
        self._check_status()

    def _set_address(self, address):
        """DFU internal command
        Set the address to read, write, execute"""
        self.log.debug("Set Adress: %x", address)
        ret = self._cmd_out(_DFUComand.DFU_DNLOAD, 0, struct.pack("<BI",0x21, address))
        self._check_status()

    def _read_mem(self, length):
        """DFU internal command
        Reads length number of bytes from memory.
        The address is set by the set_address() command."""
        self.log.debug("Reading memory length: %d", length)

        ret = self._cmd_in(_DFUComand.DFU_UPLOAD, 2, length)

        self.log.debug("Data receving: %s", "".join(["{:02X}".format(i) for i in ret]))
        return ret, self._check_status()

    def _get_cmd(self):
        """DFU internal command
        Requests a list of supportet commands from the DFU device"""
        raise NotImplementedError()

    def _write_mem(self):
        """DFU internal command
        """
        raise NotImplementedError()

    def _erease(self):
        """DFU internal command
        """
        raise NotImplementedError()

    def _read_unprotect(self):
        """DFU internal command
        """
        raise NotImplementedError()


    def _leave_dfu(self):
        """DFU internal command
        Executes the code at the address set by set_address()"""
        self.log.debug("Leaving DFU mode")

        ret = self._cmd_out(_DFUComand.DFU_DNLOAD, 0, b"")
        status = self._check_status()

        if status["bState"] != _bState.dfuMANIFEST:
            self.log.error("Leave dfu command faild")
            raise DFUException(status)

    def read_at_addr_len(self, addr, length):
        """Read data from the DFU device.

        Arguments:
            addr   -- address to the first byte to read
            length -- number of byte to read

        This will not work for all addresses and reading might be not allowed by the bootloader.

        Returns the read data as bytearray."""
        try:
            self._clear_status()
        except:
            # needs to be done twice to clear the error
            self._clear_status()

        self._abort()

        ## Test
        self._set_address(addr)
        self._check_status()
        self._abort()
        self._check_status()
        ret, _ = self._read_mem(length)

        self._abort()
        return ret

    def get_uuid(self):
        """Returns the UUID of an STM32F0 as bytearray.

        This might not work on some targets as it does a read outside the memory region
        and the bootloader might not allow this.
        """
        uuid = self.read_at_addr_len(0x1FFFF7AC, 12)
        uuid = bytearray(uuid)

        self.log.debug("Get UUID: %s", uuid.hex())

        return uuid

    def enter_user_code(self):
        """Executes the user application in flash.
        This can be used to exit the DFU bootloader.

        The code reads the entry point from flash and jumps to it.
        """
        self.log.debug("Entering user code")

        try:
            self._clear_status()
        except:
            # needs to be done twice to clear the error
            self._clear_status()

        self._abort()

        ret = self.read_at_addr_len(0x08000000, 8)
        addr = struct.unpack("I", ret[4:8])[0]
        self.log.debug("Code entry address: %x", addr)

        self._abort()


        self._set_address(addr)

        self._get_status()

        self._leave_dfu()

class DfuUtilNotFoundError(Exception):
    pass

class DfuUtilFailedError(Exception):
    pass

def dfu_util_version():
    """
    Get a dfu-util version string like "dfu-util 0.10" or raise an Exception when dfu-util can not be found
    """

    try:
        output = subprocess.check_output([DFU_UTIL_CMD, "-V"])
        output = output.decode('utf-8') + '\n'
        dfu_version, *_ = output.split('\n')

        return dfu_version
    except FileNotFoundError:
        raise DfuUtilNotFoundError("dfu-util not found. Might not be installed.")

def dfu_util_flash_firmware(firmware_path, usb_path):
    """Flash firmware to USB-Mux in DFU mode.
    This uses the command line tool dfu-util so that must be installed.

    Arguments:
    firmware_path -- Path to the firmware file as string
    usb_path      -- USB path to USB device (example: 1-2.2.1)

    Throws an Exception if dfu-util is not installed oder dfu-util failed"""
    try:
        res = subprocess.run([DFU_UTIL_CMD,
            "-d", "0483:df11",
            "-a", "0",
            "-D", firmware_path,
            "-s", "0x8000000",
            "--path", usb_path])
        if res.returncode != 0:
            raise DfuUtilFailedError("dfu-util failed with: {}".format(res))
    except FileNotFoundError:
        raise DfuUtilNotFoundError("dfu-util not found. Might not be installed.")

def dfu_util_flash_config(file_path, usb_path):
    """Flash config to USB-Mux in DFU mode.
    This uses the command line tool dfu-util so that must be installed.

    Arguments:
    firmware_path -- Path to the config file as string
    usb_path      -- USB path to USB device (example: 1-2.2.1)

    Throws an Exception if dfu-util is not installed oder dfu-util failed"""
    try:
        res = subprocess.run([DFU_UTIL_CMD,
            "-d", "0483:df11",
            "-a", "0",
            "-D", file_path,
            "-s", "0x8007c00",
            "--path", usb_path])
        if res.returncode != 0:
            raise Exception("dfu-util failed with: {}".format(res))
    except FileNotFoundError:
        raise Exception("dfu-util not found. Might not be installed.")

def _find_dfu_device(search_path):
    try:
        dfu = DFU(path=search_path)
    except ToManyDFUDeviceFound as e:
        print("Found more then one DFU Device.")
        print("Please provide a path to the DFU Device you want to use")
        for path in e.paths:
            print("  * {}".format(path))
        exit(1)

    except NoDFUDeviceFound:
        print("No DFU device found")
        exit(2)

    return dfu

def _flash_firmware(args):
    dfu = _find_dfu_device(args.path)

    if args.file is None:
        print("No firmware file specified")
        exit(3)

    dfu_util_flash_firmware(args.file, dfu.get_path())


def _flash_config(args):
    dfu = _find_dfu_device(args.path)

    if args.file is None:
        print("No config file specified")
        exit(3)

    dfu_util_flash_config(args.file, dfu.get_path())

def _update_firmware(args):
    mux_list = Mux.find_devices()

    if args.file is None:
        print("No config file specified")
        exit(3)

    selected_mux = None
    for mux in mux_list:
        if args.path == mux["path"]:
            selected_mux = mux
            break
        if args.serial == mux["serial"]:
            selected_mux = mux
            break

    print("Updating USBMux {} @ {}".format(selected_mux["serial"], selected_mux["path"]))

    path = selected_mux["path"]
    mux = Mux(path=path)
    mux.enter_dfu()
    sleep(1)

    dfu = _find_dfu_device(path)

    dfu_util_flash_firmware(args.file, dfu.get_path())

    dfu.enter_user_code()

def _leave_dfu(args):
    dfu = _find_dfu_device(args.path)
    dfu.enter_user_code()

if __name__ == "__main__":

    commands = {
            "flash_firmware": _flash_firmware,
            "flash_config": _flash_config,
            "update": _update_firmware,
            "leave_dfu": _leave_dfu,
            }

    parser = argparse.ArgumentParser("can_isp.py")
    parser.add_argument(
        "function",
        help="Function to perform",
        choices=commands,
    )
    parser.add_argument(
        "--file",
        "-f",
        help="File to use as flash or config",
    )
    parser.add_argument(
        "--path",
        "-p",
        help="USB Path to USB device (1-4.1)",
    )
    parser.add_argument(
        "--serial",
        "-s",
        help="USBMux serial numer (00001.00020)",
    )
    parser.add_argument(
        "-v",
        help="Be verbose",
        action="store_true",
    )
    args = parser.parse_args()

    commands[args.function](args)




#!/usr/bin/env python3

# SPDX-License-Identifier: LGPL-2.1-or-later

# Copyright (C) 2021 Pengutronix, Chris Fiege <entwicklung@pengutronix.de>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

import argparse
import errno
from .usbmuxctl import Mux, UmuxNotFound, NoPriviliges
import json
import sys
import termcolor
import usb.core
from .update import DfuUtilFailedError, DfuUtilNotFoundError

class ConnectionNotPossible(Exception):
    pass

artwork = {}
artwork["DUT-Host Device-Host"] = """                                     +-----------------------+
                                     | USB-Mux               |
                                  +--|                       |
                                  |  | SN:   {:11s}     |
                                  |  | Path: {:16s}|
                                  |  +-----------------------+
       VCC: {:1.2f}V    +---------+  |
Host |>--------------|       1 |--+         ID: {}
                     |         |           VCC: {:1.2f}V
                     |  {:6} |---------------------|> DUT
                     |         |
                     |       3 |---------------------|> Device
                     +---------+           VCC: {:1.2f}V"""

artwork["None"] = """                                     +-----------------------+
                                     | USB-Mux               |
                                  +--|                       |
                                  |  | SN:   {:11s}     |
                                  |  | Path: {:16s}|
                                  |  +-----------------------+
       VCC: {:1.2f}V    +---------+  |
Host |>--------------|       1 |--+         ID: {}
                     |         |           VCC: {:1.2f}V
                     |  {:6} |----x    ------------|> DUT
                     |         |
                     |       3 |----x    ------------|> Device
                     +---------+           VCC: {:1.2f}V"""

artwork["DUT-Host"] = """                                     +-----------------------+
                                     | USB-Mux               |
                                  +--|                       |
                                  |  | SN:   {:11s}     |
                                  |  | Path: {:16s}|
                                  |  +-----------------------+
       VCC: {:1.2f}V    +---------+  |
Host |>--------------|       1 |--+         ID: {}
                     |         |           VCC: {:1.2f}V
                     |  {:6} |---------------------|> DUT
                     |         |
                     |       3 |----x    ------------|> Device
                     +---------+           VCC: {:1.2f}V"""

artwork["Device-Host"] = """                                     +-----------------------+
                                     | USB-Mux               |
                                  +--|                       |
                                  |  | SN:   {:11s}     |
                                  |  | Path: {:16s}|
                                  |  +-----------------------+
       VCC: {:1.2f}V    +---------+  |
Host |>--------------|       1 |--+         ID: {}
                     |         |           VCC: {:1.2f}V
                     |  {:6} |----x    ------------|> DUT
                     |         |
                     |       3 |---------------------|> Device
                     +---------+           VCC: {:1.2f}V"""


artwork["DUT-Device"] = """                                     +-----------------------+
                                     | USB-Mux               |
                                  +--|                       |
                                  |  | SN:   {:11s}     |
                                  |  | Path: {:16s}|
                                  |  +-----------------------+
       VCC: {:1.2f}V    +---------+  |
Host |>--------------|       1 |--+         ID: {}
                     |         |           VCC: {:1.2f}V
                     |  {:6} |----x    +-----------|> DUT
                     |         |         |
                     |       3 |----x    +-----------|> Device
                     +---------+           VCC: {:1.2f}V"""

raw_status_annotations={
    "data_links": "List of active links of USB data lines",
    "device": "The device this status message is printed for",
    "dut_otg_input": "Measured level on the ID pin on the DUT port",
    "dut_otg_output": "State of the open-drain MOSFET on the ID pin on the DUT port",
    "dut_power_lockout": "State of the 'Lock' switch on the USB-Mux",
    "power_links": "List of active links of USB power lines",
    "voltage_device": "Measured voltage on the device port",
    "voltage_dut": "Measured voltage on the DUT port",
    "voltage_host": "Measured voltage on the host port",
}

def _error_and_exit(error_message, rc=1):
    print(
        termcolor.colored(error_message, "red"),
        file=sys.stderr,
    )
    exit(rc)


def _ui_messages(status):
    """
    Generates a list of messages that should be displayed.
    """

    messages_list = []
    if not status["device"]["sw_up_to_date"]:
        messages_list.append("Software update for USB-Mux available")
    if status["voltage_host"] < 4.5:
        messages_list.append(
            termcolor.colored(
                "WARN: Host USB voltage is very low ({:0.1f}V)!".format(status["voltage_host"]),
                "red", attrs=['reverse']
            )
        )
    if status["voltage_host"] > 5.3:
        messages_list.append(
            termcolor.colored(
                "WARN: Host USB voltage is very high ({:0.1f}V)!".format(status["voltage_host"]),
                "red", attrs=['reverse']
            )
        )
    if status["voltage_dut"] > 5.3:
        messages_list.append(
            termcolor.colored(
                "WARN: DUT USB voltage is very high ({:0.1f}V)!".format(status["voltage_host"]),
                "red", attrs=['reverse']
            )
        )
    messages = " ".join(messages_list)
    return messages

def list_usb(args):
    if args.json:
        res = {
            "command": "list",
            "list": Mux.find_devices(),
            "error": False,
        }
        print(json.dumps(res))
    else:
        print("Serial      | USB-Path           | Host-DUT Lock? | Connections")
        print("----------- | ------------------ | -------------- | -----------")
        for d in Mux.find_devices():
            mux = Mux(path=d["path"])
            status = mux.get_status()
            if status["data_links"] == "":
                connections = "None"
            else:
                connections = status["data_links"]
            lock = "locked" if status["dut_power_lockout"] == True else "unlocked"

            messages = _ui_messages(status)
            print("{:11} | {:18} | {:14} | {:14} {}".format(
                d["serial"],
                d["path"],
                lock,
                connections,
                messages,
            ))


def show_status(status, raw=False):
    if raw:
        for k,v in sorted(status.items()):
            if k in raw_status_annotations:
                print("# {}".format(raw_status_annotations[k]))
            print("{}: {}".format(k,v))
            print()
    else:
        print(artwork[status["data_links"]].format(
            status["device"]["serial_number"],
            status["device"]["usb_path"],
            status["voltage_host"],
            "High" if status["dut_otg_input"] else "Low",
            status["voltage_dut"],
            "LOCKED" if status["dut_power_lockout"] else "     2",
            status["voltage_device"],
        ))
        messages = _ui_messages(status)
        print(messages)

def find_umux(args):
    mux = Mux(serial_number=args.serial, path=args.path)
    return mux

def status(args):
    result = {
        "command": "status",
        "error": False,
    }
    try:
        mux = find_umux(args)
        status = mux.get_status()
        result["status"] = status
    except UmuxNotFound as e:
        result["error"] = True
        result["errormessage"] = "Failed to find the defined USB-Mux"

    if args.json:
        print(json.dumps(result))
    else:
        if result["error"]:
            _error_and_exit("Failed to connect to device: {}".format(result["errormessage"]))
        else:
            show_status(result["status"], args.raw)

def disconnect(args):
    result = {
        "command": "disconnect",
        "error": False,
    }
    try:
        mux = find_umux(args)
        id_pull_low = None if args.no_id else False

        mux.connect("None", id_pull_low)

        result["status"] = mux.get_status()
    except UmuxNotFound as e:
        result["error"] = True
        result["errormessage"] = "Failed to find the defined USB-Mux"
    except ConnectionNotPossible as e:
        result["error"] = True
        result["errormessage"] = str(e)

    if args.json:
        print(json.dumps(result))
    else:
        if result["error"]:
            _error_and_exit("Failed to set connection: {}".format(result["errormessage"]))
        else:
            show_status(result["status"], args.raw)

def connect(args):
    result = {
        "command": "connect",
        "error": False,
    }
    try:
        mux = find_umux(args)
        state = mux.get_status()
        links = []
        id_pull_low = False
        if args.host_dut:
            links.append("DUT-Host")
            if state["dut_power_lockout"] == True:
                raise ConnectionNotPossible(
                    "DUT-to-host connection is locked in hardware. "+\
                    "Refusing to set connection. "+\
                    "Maybe set 'Lock' switch in the other position?")
        if args.host_device:
            links.append("Device-Host")
        if args.dut_device:
            links.append("DUT-Device")
            id_pull_low = True

        if len(links) == 0:
            links = ["None"]

        if args.no_id:
            id_pull_low = None
        mux.connect(" ".join(links), id_pull_low)

        result["status"] = mux.get_status()
    except UmuxNotFound as e:
        result["error"] = True
        result["errormessage"] = "Failed to find the defined USB-Mux"
    except ConnectionNotPossible as e:
        result["error"] = True
        result["errormessage"] = str(e)

    if args.json:
        print(json.dumps(result))
    else:
        if result["error"]:
            _error_and_exit("Failed to set connection: {}".format(result["errormessage"]))
        else:
            show_status(result["status"], args.raw)

def id(args):
    result = {
        "command": "id",
        "error": False,
    }
    try:
        mux = find_umux(args)

        if args.float:
            state = False
        if args.pull_low:
            state = True

        mux.pull_otg_id_low(state)

        result["status"] = mux.get_status()
    except UmuxNotFound as e:
        result["error"] = True
        result["errormessage"] = "Failed to find the defined USB-Mux"

    if args.json:
        print(json.dumps(result))
    else:
        if result["error"]:
            _error_and_exit("Failed to connect to device: {}".format(result["errormessage"]))
        else:
            show_status(result["status"], args.raw)

def dfu(args):
    result = {
        "command": "dfu",
        "error": False,
    }
    try:
        mux = find_umux(args)
        mux.enter_dfu()
    except UmuxNotFound as e:
        result["error"] = True
        result["errormessage"] = "Failed to find the defined USB-Mux"

    if args.json:
        print(json.dumps(result))
    else:
        if result["error"]:
            _error_and_exit("Failed to connect to device: {}".format(result["errormessage"]))
        else:
            print("OK")

def software_update(args):
    result = {
        "command": "software_update",
        "error": False,
    }

    # Make sure there is a USB-Mux selected
    if args.serial is None and args.path is None:
        result["error"] = True
        result["errormessage"] = "No serial number or path"
    else:
        try:
            mux = find_umux(args)
            mux.update_software()
        except UmuxNotFound as e:
            result["error"] = True
            result["errormessage"] = "Failed to connect to device: Failed to find the defined USB-Mux"
        except DfuUtilNotFoundError:
            result["error"] = True
            result["errormessage"] = "Could not find tool 'dfu-util'. Please install using your package manager and re-run this command."
        except DfuUtilFailedError as e:
            result["error"] = True
            result["errormessage"] = "'dfu-util' failed: '{}'. Please check the log above for hints how to fix this.".format(e)
        except usb.core.USBError as err:
            if err.errno == errno.EACCES:
                result["error"] = True
                result["errormessage"] = "'dfu-util' failed. This probably happend because of " +\
                                        "insufficient permissions: {} ".format(err) +\
                                        "Disconnect and reconnect the USB-Mux."
            else:
                result["error"] = True
                result["errormessage"] = "Unhandled USBError: {}".format(err)

    if args.json:
        print(json.dumps(result))
    else:
        if result["error"]:
            _error_and_exit(result["errormessage"])
        else:
            print("OK")

def main():
    parser = argparse.ArgumentParser(description='USB-Mux control')
    parser.add_argument('--serial',
                       help='Serial number of the USB-Mux')
    parser.add_argument('--path',
                       help='path to the USB-Mux')
    format_parser = parser.add_mutually_exclusive_group()
    format_parser.add_argument('--json',
                        help="Format output as json. Useful for scripting.",
                        action="store_true")
    format_parser.add_argument('--raw',
                               help="Format output as raw info. Useful for command line scripting.",
                               action="store_true")

    subparsers = parser.add_subparsers(help='Supply one of the following commands to interact with the USB-Mux')
    subparsers.required = True
    subparsers.dest = 'command'

    # list subcommand
    parser_list = subparsers.add_parser('list', help='Lists all connected USB-Mux')
    parser_list.set_defaults(func=list_usb)

    # status subcommand
    parser_status = subparsers.add_parser('status', help='Get the status of a USB-Mux')
    parser_status.set_defaults(func=status)

    # update subcommand
    parser_update = subparsers.add_parser('update', help='Update software on the USB-Mux')
    parser_update.set_defaults(func=software_update)

    # disconnect/connect subcommands and arguments
    parser_disconnect = subparsers.add_parser('disconnect', help='Clear all connections between the ports of the USB-Mux')
    parser_disconnect.set_defaults(func=disconnect)

    parser_connect = subparsers.add_parser('connect', help='Make connections between the ports of the USB-Mux')
    parser_connect.set_defaults(func=connect)
    connect_group = parser_connect.add_argument_group()
    connect_group.required = True
    parser_connect.add_argument('--dut-device', help='Connect DUT and Device', action='store_true')
    parser_connect.add_argument('--host-dut', help='Connect Host and DUT', action='store_true')
    parser_connect.add_argument('--host-device', help='Connect Host and Device', action='store_true')

    for subparser in (parser_connect, parser_disconnect):
        subparser.add_argument(
            '--no-id',
            help="Do not change ID pin if DUT-Port is switched. "+\
            "Allows to switch the ID pin independent of the connections.",
            action="store_true",
        )

    # id subcommand and arguments
    parser_id = subparsers.add_parser('id', help='Set the state of the ID-Pin to the DUT')
    parser_id.set_defaults(func=id)

    group_id= parser_id.add_mutually_exclusive_group()
    group_id.required = True
    group_id.add_argument('--float', help='Let the ID-Pin float', action="store_true")
    group_id.add_argument('--pull-low', help='Pull the ID-pin low', action="store_true")

    # DFU subcommand (enter dfu without acutally performing the flashing)
    parser_dfu = subparsers.add_parser('dfu', help='Send the USB-Mux into the USB-Bootloader mode.')
    parser_dfu.set_defaults(func=dfu)

    args = parser.parse_args()
    args.func(args)

if __name__ == "__main__":
    main()



#!/usr/bin/env python3

# SPDX-License-Identifier: LGPL-2.1-or-later

# Copyright (C) 2021 Pengutronix, Chris Fiege <entwicklung@pengutronix.de>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

import usb.core
from time import sleep
from sys import stderr
from .firmware import version
import errno

def path_from_usb_dev(dev):
    """Takes an pyUSB device as argument and returns a string.
    The string is a Path representation of the position of the USB device on the USB bus tree.

    This path is used to find a USB device on the bus or all devices connected to a HUB.
    The path is made up of the number of the USB controller followed be the ports of the HUB tree."""
    dev_path = ".".join([str(i) for i in dev.port_numbers])
    dev_path = "{}-{}".format(dev.bus, dev_path)
    return dev_path

class UmuxNotFound(Exception):
    pass

class NoPriviliges(Exception):
    pass

class ProtocollVersionMissmatch(Exception):
    pass

class Mux():
    """
    This class implements a driver for a USB-Mux device.
    """

    # USB-Commands
    _GET_STATUS = 0
    _SET_POWER = 1
    _SET_DATA = 2
    _SET_OTG  = 3
    _DFU = 42
    _SW_VERSION = 254
    _PROTO_VERSION = 255

    _USB_IDs = [
        {
            # This is our actual Linux Automation GmbH Vendor ID and our actual Product ID.
            "VENDORID":  0x33f7,
            "PRODUCTID": 0x0001,
        },
        {
            # This is a _fake_ Vendor-ID and Product-ID that we have choosen by throwing dice during
            # development. This ID is only the the tansition phase. There should be no USB-Muxes
            # in the wild using this ID!
            "VENDORID":  0x5824,
            "PRODUCTID": 0x27dd,
        },
    ]

    # ADC calibration data
    ADC_RANGE = (1<<16)-1
    ADC_SCALE = 3.3*3

    # Possible Links that can be set by the USB-Mux on power and data links
    # The values of this dict are used for the connect() method.
    _LINKS = {0: "None", 1: "Device-Host", 2: "DUT-Host", 3: "DUT-Host Device-Host", 4: "DUT-Device"}

    @staticmethod
    def find_devices():
        """
        Searches for a list of USB-Mux devices.

        Returns a list of all found devices:
        [{"serial": <SN>, "path": <UsbPath>}, ...]
        """
        found = []
        for ids in Mux._USB_IDs:
            devices = usb.core.find(
                find_all=True,
                idVendor=ids["VENDORID"],
                idProduct=ids["PRODUCTID"]
            )
            for dev in devices:
                path = path_from_usb_dev(dev)
                found.append({
                    "serial": dev.serial_number,
                    "path": path,
                })
        return found

    def __init__(self, serial_number=None, path=None):
        """
        Create a new Mux instance.

        Selects the device by the given serial_numer or USB path.
        Both values can be obtained using Mux.find_devices().
        If both are given serial_number has priority over path.
        If neither is given the first found USB-Mux will be used.
        """
        def find_filter(dev):
            dev_path = path_from_usb_dev(dev)

            # Perform a dummy-acces on the device to make sure
            # we actually have permissions to access it, as
            # otherwise accessing dev.serial_number will result
            # in an exception being thrown from which the reason
            # of the error is not immediately obvious.
            try:
                dev.get_active_configuration()
            except usb.core.USBError as e:
                if e.errno == errno.EACCES:
                    print("Access denied while checking serial number for device:",
                          dev_path, file=stderr)

                    return False

                raise e

            if not serial_number is None:
                return dev.serial_number == serial_number
            if not path is None:
                return dev_path == path
            return True

        for ids in Mux._USB_IDs:
            self._dev = usb.core.find(
                idVendor=ids["VENDORID"],
                idProduct=ids["PRODUCTID"],
                custom_match=find_filter,
            )
            if self._dev is not None:
                break

        if self._dev is None:
            raise UmuxNotFound()

        try:
            # check if we support the protocol version reported by the usbmux
            proto_version = self._send_cmd(self._PROTO_VERSION)
            if len(proto_version) != 8:
                raise ProtocollVersionMissmatch("The protocol version reported by the USB-Mux is not supported by this control tool.")
            self._proto_version = "".join([str(x) for x in proto_version])
            if self._proto_version not in ["00000000"]:
                raise ProtocollVersionMissmatch("The protocol version reported by the USB-Mux is not supported by this control tool.")
        except ValueError:
            raise NoPriviliges("Could not communicate with USB-device. Check privileges, maybe add udev-rule")

        # read sw version
        self.sw_version = str(self._send_cmd(self._SW_VERSION), "utf-8")
        self.sw_version_num = version.version_from_string(self.sw_version)

    def _send_cmd(self, cmd, arg=0):
        """
        Sends a low level USB control transfer to the device.
        """
        data = self._dev.ctrl_transfer((1<<7) | (2<<5) | 0,     0xff, cmd, arg, 128)
        return data

    def _parse_return(self, pkg):
        """
        Parses the status returned by most USB commands.

        Returns a dict with the available information on the hardware.
        """
        if len(pkg) != 8:
            raise Exception("Invalid Package length")

        path = ".".join([str(i) for i in self._dev.port_numbers])
        path = "{}-{}".format(self._dev.bus, path)
        state = {
            "voltage_host": \
                (pkg[0]<<8 | pkg[1]) * Mux.ADC_SCALE / Mux.ADC_RANGE,
            "voltage_device": \
                (pkg[2]<<8 | pkg[3]) * Mux.ADC_SCALE / Mux.ADC_RANGE,
            "voltage_dut": \
                (pkg[4]<<8 | pkg[5]) * Mux.ADC_SCALE / Mux.ADC_RANGE,
            "dut_power_lockout": \
                (pkg[6] & 1) != 0,
            "dut_otg_output": \
                (pkg[6] & 2) != 0,
            "dut_otg_input": \
                (pkg[7] & 128) != 0,
            "power_links": \
                Mux._LINKS[ (pkg[6]>>2)&0b111 ],
            "data_links": \
                Mux._LINKS[ (pkg[6]>>5)&0b111 ],
            "device": {
                "usb_path": path,
                "serial_number": self._dev.serial_number,
                "product_name": self._dev.product,
                "sw_version": self.sw_version,
                "sw_up_to_date": self.is_software_up_to_date(),
            },
        }
        return state

    def _connect_power(self, num):
        """
        Applies a connection for the power links on the USB-Mux.
        num must be one of Mux._LINKS.

        If the Host-DUT link is locked in hardware this command will
        still succeed, but this link will not be set. The state returned
        will reflect this.

        Return the bytefiled received by the USB-Mux.
        This can be parsed using self._parse_return()
        """
        if 4 <= num <= 0:
            raise Exception("{} is not a valid power connection id".format(num))
        data = self._send_cmd(self._SET_POWER, num)
        return self._parse_return(data)

    def _connect_data(self, num):
        """
        Applies a connection for the data links on the USB-Mux.
        num must be one of Mux._LINKS.

        If the Host-DUT link is locked in hardware this command will
        still succeed, but this link will not be set. The state returned
        will reflect this.

        Return the bytefiled received by the USB-Mux.
        This can be parsed using self._parse_return()
        """
        if 4 <= num <= 0:
            raise Exception("{} is not a valid data connection id".format(num))
        data = self._send_cmd(self._SET_DATA, num)
        return self._parse_return(data)

    def pull_otg_id_low(self, state):
        """
        Sets the state of the ID pin on the DUT port.

        If state is True: Pulls ID pin low
        If state if False: Leave ID pin floating, a 100k Pull Up is active

        Returns a dict with the state reported by the hardware.
        """
        if state == True:
            id_state = 1
        elif state == False:
            id_state = 0
        else:
            raise Exception("{} is not a valid data for otg_id".format(state))
        data = self._send_cmd(self._SET_OTG, id_state)
        return self._parse_return(data)

    def get_status(self):
        """
        Queries the state of the USB-Mux.

        Returns a dict with the state reported by the hardware.
        """
        data = self._send_cmd(self._GET_STATUS)
        return self._parse_return(data)

    def enter_dfu(self):
        """
        Disconnects all Links and resets the CPU into the DFU-Mude.
        DFU-Mode is provided by the ROM Code.
        This mode is used to transfer firmware onto the device.

        To resume normal operation the USB-Mux must be either reset
        using the DFU-Mode or power cycled.
        """
        self._connect_power(0)
        self._connect_data(0)

        sleep(0.1)
        try:
            self._send_cmd(self._DFU)
        except usb.core.USBError:
            pass

    def connect(self, links, id_pull_low = None):
        """
        Applies a connection between ports of the USB-Mux.

        Before switching to a new connection all current connections
        are being removed and this methods gives the USB-Mux about 0.5 s to settle.
        This will cycle the power to all connected USB devices.
        Afterwards the new links are being connected.

        links must be one of MUX._LINKS.

        id_pull_low affects the ID pin on the DUT port:
        if id_pull_low is None: The ID pin is not altered at all. Use
                                 pull_otg_low() to set the ID pin independent
                                 of the USB links.
        if id_pull_low is True: The ID pin is set floating when the previous connections
                                are being disconnected. The ID pin will be pulled low before
                                the new links are being connected.
        if id_pull_low is False: The ID pin is set floating when the previous connections
                                 are being disconnected. The ID pin will be left floating
                                 when the new links are being connected.
        """
        num = None
        for key, value in Mux._LINKS.items():
            if value == links:
                num = key
        if num is None:
            raise Exception("Invalid connection {}".format(links))

        self._connect_power(0)
        self._connect_data(0)
        if not id_pull_low == None:
            self.pull_otg_id_low(False)

        sleep(0.5) # Gives switches time to settle and devices to power off

        self._connect_power(num)
        self._connect_data(num)
        if not id_pull_low == None:
            self.pull_otg_id_low(id_pull_low)

        sleep(0.3) # Wait a little moment for switches to settle

    def __str__(self):
        path = path_from_usb_dev(dev)
        path = "Connected to:\n- ID:   {}\n- Path: {}\n- Name: {}".format(self._dev.serial_number, path, self._dev.product)
        return path

    def is_software_up_to_date(self):
        return version.FIRMWARE_VERSION <= self.sw_version_num

    def update_software(self):
        """Updates the usbmux software"""

        from .update import (
            DFU,
            dfu_util_flash_firmware,
            dfu_util_version,
        )

        # We do not care about the actual version string but want this
        # to fail early (before we've entered DFU mode) if dfu-util is not installed.
        _ = dfu_util_version()
        self.enter_dfu()
        sleep(1)

        usb_path = path_from_usb_dev(self._dev)

        dfu = DFU(path=usb_path)

        dfu_util_flash_firmware(version.FIRMWARE_FILE, dfu.get_path())

        dfu.enter_user_code()



#!/usr/bin/env python3

import fastentrypoints

from setuptools import setup

setup(
    include_package_data=True,
    name="usbmuxctl",
    version="0.1.2",
    author="Chris Fiege",
    author_email="python@pengutronix.de",
    license="LGPL-2.1-or-later",
    url="https://github.com/linux-automation/usbmuxctl",
    description="Tool to control an USB-Mux from the command line",
    packages=['usbmuxctl'],
    install_requires=[
        'pyusb',
        'termcolor',
    ],
    entry_points={
        'console_scripts': [
            'usbmuxctl = usbmuxctl.__main__:main',
        ],
    },
    classifiers=[
        "License :: OSI Approved :: GNU Lesser General Public License v2 or later (LGPLv2+)"
    ]
)
